/*
 * livepatch_bsc1196959
 *
 * Fix for CVE-2021-39698, bsc#1196959
 *
 *  Upstream commits:
 *  42288cb44c4b ("wait: add wake_up_pollfree()")
 *  a880b28a71e3 ("binder: use wake_up_pollfree()")
 *  9537bae0da1f ("signalfd: use wake_up_pollfree()"
 *  363bee27e258 ("aio: keep poll requests on waitqueue until completed")
 *  50252e4b5e98 ("aio: fix use-after-free due to missing POLLFREE handling")
 *  791f3465c4af ("io_uring: fix UAF due to missing POLLFREE handling")
 *
 *  SLE12-SP3 commit:
 *  not affected
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  not affected
 *
 *  SLE15-SP2 and -SP3 commit:
 *  fbcad1232a2939a59b964cee9145dbaf4fa1e78c
 *  00c1c8d55a7a4fdedb36c7993cb09c9814a4742f
 *  28c3691b228694917e6bcd8822fc16c97c7ffac7
 *  3ae391179ed2a4ee434cfbc4e932906dd8fbe8fa
 *  b026506dc6444c526bb88c0c33100b5843e60d0b
 *
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

static const struct file_operations (*klpe_signalfd_fops);

/* klp-ccp: from fs/io_uring.c */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/syscalls.h>
#include <linux/compat.h>
#include <linux/refcount.h>
#include <linux/uio.h>
#include <linux/sched/signal.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/percpu.h>
#include <linux/slab.h>
#include <linux/workqueue.h>
#include <linux/kthread.h>
#include <linux/bvec.h>
#include <net/scm.h>
#include <linux/uaccess.h>
#include <linux/nospec.h>
#include <linux/sizes.h>

struct async_list {
	spinlock_t		lock;
	atomic_t		cnt;
	struct list_head	list;

	struct file		*file;
	off_t			io_end;
	size_t			io_len;
};

struct io_ring_ctx {
	struct {
		struct percpu_ref	refs;
	} ____cacheline_aligned_in_smp;

	struct {
		unsigned int		flags;
		bool			compat;
		bool			account_mem;

		/* SQ ring */
		struct io_sq_ring	*sq_ring;
		unsigned		cached_sq_head;
		unsigned		sq_entries;
		unsigned		sq_mask;
		unsigned		sq_thread_idle;
		unsigned		cached_sq_dropped;
		struct io_uring_sqe	*sq_sqes;

		struct list_head	defer_list;
	} ____cacheline_aligned_in_smp;

	/* IO offload */
	struct workqueue_struct	*sqo_wq;
	struct task_struct	*sqo_thread;	/* if using sq thread polling */
	struct mm_struct	*sqo_mm;
	wait_queue_head_t	sqo_wait;
	struct completion	sqo_thread_started;

	struct {
		/* CQ ring */
		struct io_cq_ring	*cq_ring;
		unsigned		cached_cq_tail;
		atomic_t		cached_cq_overflow;
		unsigned		cq_entries;
		unsigned		cq_mask;
		struct wait_queue_head	cq_wait;
		struct fasync_struct	*cq_fasync;
		struct eventfd_ctx	*cq_ev_fd;
	} ____cacheline_aligned_in_smp;

	/*
	 * If used, fixed file set. Writers must ensure that ->refs is dead,
	 * readers must ensure that ->refs is alive as long as the file* is
	 * used. Only updated through io_uring_register(2).
	 */
	struct file		**user_files;
	unsigned		nr_user_files;

	/* if used, fixed mapped user buffers */
	unsigned		nr_user_bufs;
	struct io_mapped_ubuf	*user_bufs;

	struct user_struct	*user;

	const struct cred	*creds;

	struct completion	ctx_done;

	struct {
		struct mutex		uring_lock;
		wait_queue_head_t	wait;
	} ____cacheline_aligned_in_smp;

	struct {
		spinlock_t		completion_lock;
		bool			poll_multi_file;
		/*
		 * ->poll_list is protected by the ctx->uring_lock for
		 * io_uring instances that don't use IORING_SETUP_SQPOLL.
		 * For SQPOLL, only the single threaded io_sq_thread() will
		 * manipulate the list, hence no extra locking is needed there.
		 */
		struct list_head	poll_list;
		struct list_head	cancel_list;
	} ____cacheline_aligned_in_smp;

	struct async_list	pending_async[2];

#if defined(CONFIG_UNIX)
	struct socket		*ring_sock;
#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
};

struct sqe_submit {
	const struct io_uring_sqe	*sqe;
	unsigned short			index;
	u32				sequence;
	bool				has_user;
	bool				needs_lock;
	bool				needs_fixed_file;
};

struct io_poll_iocb {
	struct file			*file;
	struct wait_queue_head		*head;
	__poll_t			events;
	bool				done;
	bool				canceled;
	struct wait_queue_entry		wait;
};

struct io_kiocb {
	union {
		struct file		*file;
		struct kiocb		rw;
		struct io_poll_iocb	poll;
	};

	struct sqe_submit	submit;

	struct io_ring_ctx	*ctx;
	struct list_head	list;
	struct list_head	link_list;
	unsigned int		flags;
	refcount_t		refs;
	unsigned long		fsize;
	u64			user_data;
	u32			result;
	u32			sequence;

	struct fs_struct	*fs;

	struct work_struct	work;
};

static void (*klpe_io_commit_cqring)(struct io_ring_ctx *ctx);

static void (*klpe_io_cqring_fill_event)(struct io_ring_ctx *ctx, u64 ki_user_data,
				 long res);

static void (*klpe_io_cqring_ev_posted)(struct io_ring_ctx *ctx);

static void (*klpe_io_put_req)(struct io_kiocb *req);

void klpp_io_poll_remove_one(struct io_kiocb *req)
{
	struct io_poll_iocb *poll = &req->poll;

	/*
	 * Fix CVE-2021-39698
	 *  +47 lines
	 */
	if (poll->file->f_op == &(*klpe_signalfd_fops)) {
		/*
		 * The wait_queue_head at poll->head might not exist
		 * anymore, and it must be made sure that
		 * signalfd_cleanup() or io_poll_wake(POLLFREE)
		 * respectively has not been invoked on it yet before
		 * accessing it.
		 *
		 * Note that once a aio poll request has been
		 * completed, either inline in io_poll_wake() or from
		 * io_poll_complete_work() scheduled therefrom, it
		 * will get removed from the ->cancel_list and can't
		 * be cancelled anymore. That is, by the mere fact of
		 * being here, we know it's not been completed yet
		 * (and won't be while here because the
		 * ctx->completion_lock is being held).
		 *
		 * Thus, the case of concern here is that
		 * io_poll_wake(POLLFREE) has been run and set
		 * req->fsize to -ECANCELED in order to record the
		 * POLLFREE event, but io_poll_complete_work() has not
		 * made it past acquisition of the
		 * ctx->completion_lock.
		 */
		if (poll->canceled)
			return;

		/*
		 * For signalfds, the wait_queue_head referenced by
		 * poll->head is from a SLAB_TYPESAFE_BY_RCU slab
		 * cache, c.f. the comment in signalfd_cleanup() and
		 * the slab cache initializtion in proc_caches_init().
		 * If req->fsize != -ECANCELED below, then
		 * io_poll_wake(POLLFREE) hasn't run yet and the RCU
		 * read side section guarantees that poll->head points
		 * to valid memory storing *some* wait_queue_head, but
		 * not necessarily the original one by the time
		 * io_poll_add()->signalfd_poll() had been executed. In
		 * either case, it is perfectly fine to take
		 * poll->head->lock.
		 */
		rcu_read_lock();
		if (READ_ONCE(req->fsize) == (unsigned long)-ECANCELED) {
			rcu_read_unlock();
			return;
		}
	}

	spin_lock(&poll->head->lock);

	/*
	 * Fix CVE-2021-39698
	 *  +10 lines
	 */
	/*
	 * Ok, got the wait_queue_head ->lock, now being fully
	 * serialized with io_poll_wake(POLLFREE).
	 */
	if (poll->file->f_op == &(*klpe_signalfd_fops) &&
	    req->fsize == (unsigned long)-ECANCELED) {
		spin_unlock(&poll->head->lock);
		rcu_read_unlock();
		return;
	}

	WRITE_ONCE(poll->canceled, true);
	if (!list_empty(&poll->wait.entry)) {
		list_del_init(&poll->wait.entry);
		queue_work(req->ctx->sqo_wq, &req->work);
	}
	spin_unlock(&poll->head->lock);

	/*
	 * Fix CVE-2021-39698
	 *  +2 lines
	 */
	if (poll->file->f_op == &(*klpe_signalfd_fops))
		rcu_read_unlock();

	list_del_init(&req->list);
}

static void klpr_io_poll_complete(struct io_ring_ctx *ctx, struct io_kiocb *req,
			     __poll_t mask)
{
	req->poll.done = true;
	(*klpe_io_cqring_fill_event)(ctx, req->user_data, mangle_poll(mask));
	(*klpe_io_commit_cqring)(ctx);
}

void klpp_io_poll_complete_work(struct work_struct *work)
{
	struct io_kiocb *req = container_of(work, struct io_kiocb, work);
	struct io_poll_iocb *poll = &req->poll;
	struct poll_table_struct pt = { ._key = poll->events };
	struct io_ring_ctx *ctx = req->ctx;
	const struct cred *old_cred;
	__poll_t mask = 0;

	old_cred = override_creds(ctx->creds);

	/*
	 * Fix CVE-2021-39698
	 *  -2 lines, +12 lines
	 */
	if (!READ_ONCE(poll->canceled)) {
		mask = vfs_poll(poll->file, &pt) & poll->events;

		if (poll->file->f_op == &(*klpe_signalfd_fops) &&
		    req->fsize == (unsigned long)-ECANCELED) {
			/*
			 * A POLLFREE/EPOLLHUP event, c.f the livepatched
			 * io_poll_wake().
			 */
			mask |= (EPOLLHUP | POLLFREE) & poll->events;
		}
	}

	/*
	 * Note that ->ki_cancel callers also delete iocb from active_reqs after
	 * calling ->ki_cancel.  We need the ctx_lock roundtrip here to
	 * synchronize with them.  In the cancellation case the list_del_init
	 * itself is not actually needed, but harmless so we keep it in to
	 * avoid further branches in the fast path.
	 */
	spin_lock_irq(&ctx->completion_lock);
	if (!mask && !READ_ONCE(poll->canceled)) {
		add_wait_queue(poll->head, &poll->wait);
		spin_unlock_irq(&ctx->completion_lock);
		goto out;
	}
	list_del_init(&req->list);
	klpr_io_poll_complete(ctx, req, mask);
	spin_unlock_irq(&ctx->completion_lock);

	(*klpe_io_cqring_ev_posted)(ctx);
	(*klpe_io_put_req)(req);
out:
	revert_creds(old_cred);
}

int klpp_io_poll_wake(struct wait_queue_entry *wait, unsigned mode, int sync,
			void *key)
{
	struct io_poll_iocb *poll = container_of(wait, struct io_poll_iocb,
							wait);
	struct io_kiocb *req = container_of(poll, struct io_kiocb, poll);
	struct io_ring_ctx *ctx = req->ctx;
	__poll_t mask = key_to_poll(key);
	unsigned long flags;

	/* for instances that support it check for an event match first: */
	if (mask && !(mask & poll->events))
		return 0;

	/*
	 * Fix CVE-2021-39698
	 *  +10 lines
	 */
	if (!mask && poll->file->f_op == &(*klpe_signalfd_fops)) {
		/*
		 * Anything but a POLLFREE/EPOLLHUP event. Don't even
		 * bother to schedule the work below, signalfd_poll()
		 * won't return anything in the context of the worker
		 * thread anyway. Leave us on the wqh so that future
		 * POLLFREE events won't get missed.
		 */
		return 0;
	}

	list_del_init(&poll->wait.entry);

	if (mask && spin_trylock_irqsave(&ctx->completion_lock, flags)) {
		list_del(&req->list);
		klpr_io_poll_complete(ctx, req, mask);
		spin_unlock_irqrestore(&ctx->completion_lock, flags);

		(*klpe_io_cqring_ev_posted)(ctx);
		(*klpe_io_put_req)(req);
	/*
	 * Fix CVE-2021-39698
	 *  +30 lines
	 */
	} else if (poll->file->f_op == &(*klpe_signalfd_fops)) {
		/*
		 * This is a POLLFREE/EPOLLHUP event from
		 * signalfd_cleanup() and the request could not get
		 * completed inline, because the spin_trylock() above
		 * failed to acquire the required ctx lock. Note that
		 * before the livepatch, the scheduled
		 * io_poll_complete_work() would have failed to
		 * extract the event via signalfd_poll() from the
		 * context of the worker thread and a UAF would have
		 * occured upon the readdition to the work_queue_head.
		 * However, simply skipping the work scheduling below
		 * alltogether to "fix" the UAF in a kind of bug
		 * compatible way won't work: io_poll_remove_one()
		 * expects the work to have been scheduled if
		 * list_empty(&poll->wait.entry) is found to be true.
		 * One approach would be to mimic cancellation,
		 * i.e. to set poll->canceled and to schedule the
		 * completion work. However, userspace might not
		 * expect this: it would look like a successful
		 * completion with an empty poll result.  Instead, do
		 * better and signal to a livepatched
		 * io_poll_complete_work() via req->fsize that there
		 * had been a POLLFREE/EPOLLHUP event and let it
		 * handle this properly. Note that req->fsize is
		 * otherwise unused for poll IO requests, in fact,
		 * it's unused for anything but write IO requests.
		 */
		WRITE_ONCE(req->fsize, (unsigned long)-ECANCELED);
		queue_work(ctx->sqo_wq, &req->work);
	} else {
		queue_work(ctx->sqo_wq, &req->work);
	}

	return 1;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1196959.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "io_commit_cqring", (void *)&klpe_io_commit_cqring },
	{ "io_cqring_ev_posted", (void *)&klpe_io_cqring_ev_posted },
	{ "io_cqring_fill_event", (void *)&klpe_io_cqring_fill_event },
	{ "io_put_req", (void *)&klpe_io_put_req },
	{ "signalfd_fops", (void *)&klpe_signalfd_fops },
};

int livepatch_bsc1196959_init(void)
{
	return __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}
